#include "cpudata.h"
#include "device.h"

#include <cstring>
#include <math.h>
#include <iostream>

#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))

#define maxRo 1e12f
#define maxT 2097152.0f
#define maxV 2048.0f
#define maxRoVisc 2048.0f
#define minRo 1e-6f
#define minT 1e-6f
#define minP 1e-6f



#define checkVal(x,minVal,maxVal) x=(min(max(((x!=x)?0.0:x),minVal),maxVal))

#define THREAD_SAFE(x)\
pthread_mutex_lock(&_mutex);\
{\
x\
}\
pthread_mutex_unlock(&_mutex)

const char * Gas::T_RO="ro";
const char * Gas::T_PR="p";
const char * Gas::T_VX="vx";
const char * Gas::T_VY="vy";
const char * Gas::T_VS="vs";
const char * Gas::T_SC="sc";
const char * Gas::T_CS="cs";
const char * Gas::T_R1="r1";
const char * Gas::T_G1="g1";
const char * Gas::T_B1="b1";
const char * Gas::T_R2="r2";
const char * Gas::T_G2="g2";
const char * Gas::T_B2="b2";
const char * Gas::T_R3="r3";
const char * Gas::T_G3="g3";
const char * Gas::T_B3="b3";
const char * Gas::T_FULL_TIME_LEFT="ft";
const char * Gas::T_MARK_TIME_LEFT="mt";

const char * CpuGravityBody::BODY_ID="id";
const char * CpuGravityBody::BODY_X="x";
const char * CpuGravityBody::BODY_Y="y";
const char * CpuGravityBody::BODY_VX="vx";
const char * CpuGravityBody::BODY_VY="vy";
const char * CpuGravityBody::BODY_MASS="m";
const char * CpuGravityBody::BODY_RADIUS="r";
const char * CpuGravityBody::BODY_GAS_V="cgv";
const char * CpuGravityBody::BODY_GAS_M="cgm";
const char * CpuGravityBody::BODY_BODY_V="cv";
const char * CpuGravityBody::BODY_LIFETIME="lt";
const char * CpuGravityBody::BODY_LIFEPATH="lp";

void Gas::check()
{
    if(1.0/sqrt(P*1.4/Ro)<0.1)
    {
        P=1;
    }

    if(Ro==Ro&&Ro>-minRo)
    {
        checkVal(Ro,0.0,maxRo);
    }
    if(P==P&&P>-minP)
    {
        checkVal(P,fabs(Ro)*minT,fabs(Ro)*maxT);
    }
    if(Ro<=minRo&&Ro>=-minRo)
    {
        Ro=0;
        P=0;
        Vx=0;
        Vy=0;
        Vs=0;
        Sc=0;
        Cs=1;
        R1=0;
        G1=0;
        B1=0;
        R2=0;
        G2=0;
        B2=0;
        R3=0;
        G3=0;
        B3=0;
        FullTimeLeft=0;
        MarkTimeLeft=0;
    }
    if(Ro>maxRo)
    {
        P=P*maxRo/Ro;
        Vs=Vs*maxRo/Ro;
        Ro=maxRo;
    }
    checkVal(Vx,-maxV,maxV);
    checkVal(Vy,-maxV,maxV);
    checkVal(Vs,0.0,fabs(Ro)*maxRoVisc);
    checkVal(Sc,0.0,4.0);
    checkVal(Cs,1e-6,1e12);
    checkVal(R1,0.0,256.0);
    checkVal(G1,0.0,256.0);
    checkVal(B1,0.0,256.0);
    checkVal(R2,0.0,256.0);
    checkVal(G2,0.0,256.0);
    checkVal(B2,0.0,256.0);
    checkVal(R3,0.0,256.0);
    checkVal(G3,0.0,256.0);
    checkVal(B3,0.0,256.0);
    checkVal(FullTimeLeft,0.0,1e36);
    checkVal(MarkTimeLeft,0.0,1e36);
}

std::vector<float> serilizeGas(std::vector<std::vector<Gas> > gas)
{
    std::vector<float> out;

    int sizex=gas.size();
    int sizey=sizex>0?gas[0].size():0;

    out.resize(2);

    memmove(&out[0],&sizex,sizeof(int));
    memmove(&out[1],&sizey,sizeof(int));

    for(int i=0;i<sizex;++i)
    {
        for(int j=0;j<sizey;++j)
        {
            out.push_back(static_cast<float>(gas[i][j].Ro));
            out.push_back(static_cast<float>(gas[i][j].P));
            out.push_back(static_cast<float>(gas[i][j].Vx));
            out.push_back(static_cast<float>(gas[i][j].Vy));
            out.push_back(static_cast<float>(gas[i][j].Vs));
            out.push_back(static_cast<float>(gas[i][j].Cs));
            out.push_back(static_cast<float>(gas[i][j].Sc));
            out.push_back(static_cast<float>(gas[i][j].R1));
            out.push_back(static_cast<float>(gas[i][j].R2));
            out.push_back(static_cast<float>(gas[i][j].R3));
            out.push_back(static_cast<float>(gas[i][j].G1));
            out.push_back(static_cast<float>(gas[i][j].G2));
            out.push_back(static_cast<float>(gas[i][j].G3));
            out.push_back(static_cast<float>(gas[i][j].B1));
            out.push_back(static_cast<float>(gas[i][j].B2));
            out.push_back(static_cast<float>(gas[i][j].B3));
            out.push_back(static_cast<float>(gas[i][j].FullTimeLeft));
            out.push_back(static_cast<float>(gas[i][j].MarkTimeLeft));
        }
    }

    return out;
}

std::vector<std::vector<Gas> > deserilizeGas(std::vector<float> input)
{
    std::vector<std::vector<Gas> > out;

    int sizex;
    int sizey;

    memmove(&sizex,&input[0],sizeof(int));
    memmove(&sizey,&input[1],sizeof(int));

    std::vector<float>::iterator it = input.begin();
    it++;
    it++;

    for(int i=0;i<sizex;++i)
    {
        for(int j=0;j<sizey;++j)
        {
            out[i][j].Ro=*(it++);
            out[i][j].P=*(it++);
            out[i][j].Vx=*(it++);
            out[i][j].Vy=*(it++);
            out[i][j].Vs=*(it++);
            out[i][j].Cs=*(it++);
            out[i][j].Sc=*(it++);
            out[i][j].R1=*(it++);
            out[i][j].R2=*(it++);
            out[i][j].R3=*(it++);
            out[i][j].G1=*(it++);
            out[i][j].G2=*(it++);
            out[i][j].G3=*(it++);
            out[i][j].B1=*(it++);
            out[i][j].B2=*(it++);
            out[i][j].B3=*(it++);
            out[i][j].FullTimeLeft=*(it++);
            out[i][j].MarkTimeLeft=*(it++);
        }
    }

    return out;
}

Gas::Gas()
{
    Ro=1.0;
    P=1.0;
    Vx=0.0;
    Vy=0.0;
    Vs=0.0;
    Sc=2.0;
    Cs=1.0;
    R1=255.0;
    G1=255.0;
    B1=255.0;
    R2=255.0;
    G2=255.0;
    B2=255.0;
    R3=255.0;
    G3=255.0;
    B3=255.0;
    FullTimeLeft=1e35;
    MarkTimeLeft=1e35;
}

float CpuGravityBody::getMass()
{
    return mass+gasRo*gasM;
}

float CpuGravityBody::getVx()
{
    return vx+gasVx*gasV;
}

float CpuGravityBody::getVy()
{
    return vy+gasVy*gasV;
}


CpuData::CpuData()
{
    pthread_mutex_init(&_mutex,NULL);
    _dt=1e-6f;
    _ndt=1;
}

CpuData::~CpuData()
{
    pthread_mutex_destroy(&_mutex);
}

void CpuData::init(std::vector<std::vector<Gas> > gas)
{
    _gas=gas;
    _time=0.0;
    _step=0;
    _frameStep=-1;
    _frameTime=-1;
}

float CpuData::getDt()
{
    return _dt;
}

void CpuData::setDt()
{

    THREAD_SAFE(
        if(_ndt<1e6)
        {
            bodyStep(_dt);
            std::cout<<_step<<" t= "<<_time<<" dt= " << _dt<<std::endl;
            _time+=_dt;
            _dt=min(_dt*2.0,_ndt);
            ++_step;
            _ndt=1e20;
        }
    );
}

void CpuData::setNextDt(float dt)
{
    THREAD_SAFE(_ndt=min(_ndt,dt););
}

float CpuData::getTime()
{
    return _time;
}

int CpuData::getStep()
{
    return _step;
}

Position CpuData::getSize()
{
    return (_gas.size()>0&&_gas[0].size()>0)?Position(_gas.size(),_gas[0].size()):Position(0,0);
}

Position CpuData::getZeroPosition()
{
    Position size = getSize();
    return Position(-size.x/2,-size.y/2);
}

std::vector<std::vector<Gas> > & CpuData::getGas()
{
    return _gas;
}

Gas CpuData::getGas(Position position)
{
    Gas gas;
    Position zero = getZeroPosition();
    position.x-=zero.x;
    position.y-=zero.y;
    if(position.x>0&&position.y>0&&position.x<_gas.size()&&position.y<_gas[0].size())
        gas=_gas[position.x][position.y];
    return gas;
}

bool CpuData::setGas(Gas gas, Position position)
{
    Position zero = getZeroPosition();
    position.x-=zero.x;
    position.y-=zero.y;
    if(position.x>=0&&position.y>=0&&position.x<_gas.size()&&position.y<_gas[0].size())
    {
        _gas[position.x][position.y]=gas;
        return true;
    }
    return false;
}

void CpuData::operator=(Device * data)
{
    data->getGpuGas(_gas);
}

bool CpuData::isFrame()
{
    return (_time>=_frameTime)||(_step>=_frameStep);
}

bool CpuData::makeFrame(float timeFrame, int enterFrame)
{
    _frameStep=_step+enterFrame;
    _frameTime=_time+timeFrame;
}

bool CpuData::addBody(CpuGravityBody body)
{
    while(_body.size()>=4096)
        removeBody(_body[0].id);

    _body.push_back(body);
}

bool CpuData::removeBody(std::string id)
{
    for(std::vector<CpuGravityBody>::iterator it=_body.begin();it!= _body.end();++it)
    {
        if((*it).id==id)
        {
            _body.erase(it);
            return true;
        }
    }
    return false;
}

struct threadBody
{
    std::vector<CpuGravityBody> * bodys;
    int shift;
    float dt;
};

float grav(float dist,float rad)
{
    if(dist>=rad)
        return 1.0/dist;
    float x=dist/rad;
    return (3*x-2*x*x)/rad;
}

bool CpuData::bodyStep(float dt)
{
    if(_body.size()<1)
        return true;

    try
    {
        std::vector<std::string> forDelete;

        for(int j=0;j<_body.size();++j)
        {
            float dx=_body[j].getVx()*dt;
            float dy=_body[j].getVy()*dt;
            float path=sqrt(dx*dx+dy*dy);
            _body[j].lifePath-=path;
            _body[j].lifeTime-=dt;
            if(_body[j].lifePath<0||_body[j].lifeTime<0)
                forDelete.push_back(_body[j].id);
        }

        for(int i=0;i<forDelete.size();++i)
            removeBody(forDelete[i]);
    }
    catch(...)
    {
        logger("exception in CpuData::bodyStep");
        return false;
    }

    return true;
}

bool CpuData::makeConstBody(GpuGravityBody * body)
{
    if(_body.size()<=4096)
    {
        for(int i=0;i<_body.size();++i)
        {
            body[i].x=_body[i].x;
            body[i].y=_body[i].y;
            body[i].radius=_body[i].radius;
            body[i].mass=_body[i].getMass();
        }
        return true;
    }
    return false;
}

int CpuData::getBodyCount()
{
    return min(4096,_body.size());
}

CpuGravityBody * CpuData::getBody(int n)
{
    if(n>=0&&n<_body.size())
    {
        return &_body[n];
    }
    return 0;
}
