#include "device.h"
#include "gpu.h"
#include "log.h"
#include <math.h>
#include <iostream>
#include <memory.h>

#define min(a,b) ((a)<(b)?(a):(b))

#define SAFE_CALL(function,function_description,description)\
error = function;\
if(error)\
{\
    std::cout<<error<<" error in "<<function_description<<" in "<<description<<std::endl;\
}

#define JOIN_THREAD(function,data,description)\
pthread_t thread;\
pthread_attr_t attr;\
int error;\
SAFE_CALL(pthread_attr_init( &attr ),"pthread_attr_init",description)\
SAFE_CALL(pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_JOINABLE ),"pthread_attr_setdetachstate",description)\
SAFE_CALL(pthread_create(&thread, &attr, function, (void *)data),"pthread_create",description)\
SAFE_CALL(pthread_join(thread,NULL),"pthread_join",description)\
SAFE_CALL(pthread_attr_destroy(&attr),"pthread_attr_destroy",description)

pthread_barrier_t BARRIER;

void * Device::thread_init(void * arg)
{
    std::pair<Device *, bool *> * input=(std::pair<Device *, bool *> *)arg;
    Device * _this = input->first;
    bool * result = input->second;
    *result=true;

    if(!cudaSelectGpu(_this->_gpuId))
    {
        logger(std::string("can't select gpu in Device::thread_init on device "), &_this->_gpuId);
        *result = false;
        pthread_exit(NULL);
    }

    _this->_gpuData.reset(new GpuData());
    _this->_cpuBufer.reset(new GpuData());
    _this->_upMark.reset(new GpuData());
    _this->_dwMark.reset(new GpuData());

    _this->_transponate=false;
    int size=_this->_cpuData->getSize().x*(_this->getHeight()+2);

    try
    {
        _this->_gpuData->Ro0=cudaAllocateMemory(size);
        _this->_gpuData->Ro1=cudaAllocateMemory(size);
        _this->_gpuData->P0=cudaAllocateMemory(size);
        _this->_gpuData->P1=cudaAllocateMemory(size);
        _this->_gpuData->Vx0=cudaAllocateMemory(size);
        _this->_gpuData->Vx1=cudaAllocateMemory(size);
        _this->_gpuData->Vy0=cudaAllocateMemory(size);
        _this->_gpuData->Vy1=cudaAllocateMemory(size);
        _this->_gpuData->Vs0=cudaAllocateMemory(size);
        _this->_gpuData->Vs1=cudaAllocateMemory(size);
        _this->_gpuData->Sc0=cudaAllocateMemory(size);
        _this->_gpuData->Sc1=cudaAllocateMemory(size);
        _this->_gpuData->Cs0=cudaAllocateMemory(size);
        _this->_gpuData->Cs1=cudaAllocateMemory(size);
        _this->_gpuData->R10=cudaAllocateMemory(size);
        _this->_gpuData->R11=cudaAllocateMemory(size);
        _this->_gpuData->R20=cudaAllocateMemory(size);
        _this->_gpuData->R21=cudaAllocateMemory(size);
        _this->_gpuData->R30=cudaAllocateMemory(size);
        _this->_gpuData->R31=cudaAllocateMemory(size);
        _this->_gpuData->G10=cudaAllocateMemory(size);
        _this->_gpuData->G11=cudaAllocateMemory(size);
        _this->_gpuData->G20=cudaAllocateMemory(size);
        _this->_gpuData->G21=cudaAllocateMemory(size);
        _this->_gpuData->G30=cudaAllocateMemory(size);
        _this->_gpuData->G31=cudaAllocateMemory(size);
        _this->_gpuData->B10=cudaAllocateMemory(size);
        _this->_gpuData->B11=cudaAllocateMemory(size);
        _this->_gpuData->B20=cudaAllocateMemory(size);
        _this->_gpuData->B21=cudaAllocateMemory(size);
        _this->_gpuData->B30=cudaAllocateMemory(size);
        _this->_gpuData->B31=cudaAllocateMemory(size);
        _this->_gpuData->bRo=cudaAllocateMemory(4096);
        _this->_gpuData->bVx=cudaAllocateMemory(4096);
        _this->_gpuData->bVy=cudaAllocateMemory(4096);
        _this->_gpuData->bAx=cudaAllocateMemory(4096);
        _this->_gpuData->bAy=cudaAllocateMemory(4096);

        _this->_cpuBufer->Ro0=cudaAllocateHostMemory(size);
        _this->_cpuBufer->P0=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Vx0=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Vy0=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Vs0=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Sc0=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Cs0=cudaAllocateHostMemory(size);
        _this->_cpuBufer->R10=cudaAllocateHostMemory(size);
        _this->_cpuBufer->R20=cudaAllocateHostMemory(size);
        _this->_cpuBufer->R30=cudaAllocateHostMemory(size);
        _this->_cpuBufer->G10=cudaAllocateHostMemory(size);
        _this->_cpuBufer->G20=cudaAllocateHostMemory(size);
        _this->_cpuBufer->G30=cudaAllocateHostMemory(size);
        _this->_cpuBufer->B10=cudaAllocateHostMemory(size);
        _this->_cpuBufer->B20=cudaAllocateHostMemory(size);
        _this->_cpuBufer->B30=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Ro1=cudaAllocateHostMemory(size);
        _this->_cpuBufer->P1=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Vx1=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Vy1=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Vs1=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Sc1=cudaAllocateHostMemory(size);
        _this->_cpuBufer->Cs1=cudaAllocateHostMemory(size);
        _this->_cpuBufer->R11=cudaAllocateHostMemory(size);
        _this->_cpuBufer->R21=cudaAllocateHostMemory(size);
        _this->_cpuBufer->R31=cudaAllocateHostMemory(size);
        _this->_cpuBufer->G11=cudaAllocateHostMemory(size);
        _this->_cpuBufer->G21=cudaAllocateHostMemory(size);
        _this->_cpuBufer->G31=cudaAllocateHostMemory(size);
        _this->_cpuBufer->B11=cudaAllocateHostMemory(size);
        _this->_cpuBufer->B21=cudaAllocateHostMemory(size);
        _this->_cpuBufer->B31=cudaAllocateHostMemory(size);
        _this->_cpuBufer->bRo=cudaAllocateHostMemory(4096);
        _this->_cpuBufer->bVx=cudaAllocateHostMemory(4096);
        _this->_cpuBufer->bVy=cudaAllocateHostMemory(4096);
        _this->_cpuBufer->bAx=cudaAllocateHostMemory(4096);
        _this->_cpuBufer->bAy=cudaAllocateHostMemory(4096);
        _this->_cpuBufer->gravity=cudaAllocateHostMemoryBody(4096);

        _this->_upMark->Ro0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->P0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Vx0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Vy0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Vs0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Sc0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Cs0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->R10=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->R20=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->R30=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->G10=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->G20=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->G30=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->B10=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->B20=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->B30=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Ro1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->P1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Vx1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Vy1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Vs1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Sc1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->Cs1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->R11=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->R21=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->R31=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->G11=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->G21=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->G31=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->B11=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->B21=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_upMark->B31=cudaAllocateHostMemory(_this->_cpuData->getSize().x);

        _this->_dwMark->Ro0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->P0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Vx0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Vy0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Vs0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Sc0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Cs0=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->R10=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->R20=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->R30=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->G10=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->G20=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->G30=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->B10=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->B20=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->B30=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Ro1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->P1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Vx1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Vy1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Vs1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Sc1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->Cs1=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->R11=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->R21=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->R31=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->G11=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->G21=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->G31=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->B11=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->B21=cudaAllocateHostMemory(_this->_cpuData->getSize().x);
        _this->_dwMark->B31=cudaAllocateHostMemory(_this->_cpuData->getSize().x);

    }
    catch(...)
    {
        logger(std::string("can't allocate gpu memory in Device::thread_init on device "), &_this->_gpuId);
        *result = false;
        pthread_exit(NULL);
    }

    *result = true;
    pthread_exit(NULL);
}

bool Device::init(CpuData * cpuData, Scenario * scenario, int gpuId, Position stringMarks, Device * upper, Device * downer)
{
    bool result;
    _cpuData=cpuData;
    _scenario=scenario;
    _gpuId=gpuId;
    _stringMarks=stringMarks;
    _upper=upper;
    _downer=downer;
    std::pair<Device *, bool *> input(this,&result);
    JOIN_THREAD(thread_init,&input,"Device::init")
    return result&&update();
}

void * Device::thread_update(void * arg)
{
    std::pair<Device *, bool *> * input=(std::pair<Device *, bool *> *)arg;
    Device * _this = input->first;
    bool * result = input->second;
    *result=true;

    if(!cudaSelectGpu(_this->_gpuId))
    {
        logger(std::string("can't select gpu in Device::thread_update on device "), &_this->_gpuId);
        *result = false;
        pthread_exit(NULL);
    }

    if(!_this->_gpuData.get())
    {
        logger(std::string("can't find gpu data in Device::thread_update on device "), &_this->_gpuId);
        *result = false;
        pthread_exit(NULL);
    }

    std::vector<std::vector<Gas> > & gas = _this->_cpuData->getGas();


    int index=_this->_cpuData->getSize().x;
    for(int j=_this->_stringMarks.x;j<_this->_stringMarks.y;++j)
    {
        for(int i=0;i<_this->_cpuData->getSize().x;++i)
        {
            gas[i][j].check();
            _this->_cpuBufer->Ro0[index]=(gas[i][j].Ro);
            _this->_cpuBufer->P0[index]=(gas[i][j].P);
            _this->_cpuBufer->Vx0[index]=(gas[i][j].Vx);
            _this->_cpuBufer->Vy0[index]=(gas[i][j].Vy);
            _this->_cpuBufer->Vs0[index]=(gas[i][j].Vs);
            _this->_cpuBufer->Sc0[index]=(gas[i][j].Sc);
            _this->_cpuBufer->Cs0[index]=(gas[i][j].Cs);
            _this->_cpuBufer->R10[index]=(gas[i][j].R1+0.5f);
            _this->_cpuBufer->R20[index]=(gas[i][j].R2+0.5f);
            _this->_cpuBufer->R30[index]=(gas[i][j].R3+0.5f);
            _this->_cpuBufer->G10[index]=(gas[i][j].G1+0.5f);
            _this->_cpuBufer->G20[index]=(gas[i][j].G2+0.5f);
            _this->_cpuBufer->G30[index]=(gas[i][j].G3+0.5f);
            _this->_cpuBufer->B10[index]=(gas[i][j].B1+0.5f);
            _this->_cpuBufer->B20[index]=(gas[i][j].B2+0.5f);
            _this->_cpuBufer->B30[index]=(gas[i][j].B3+0.5f);
            _this->_cpuBufer->Cs1[index]=(gas[i][j].FullTimeLeft);
            _this->_cpuBufer->Sc1[index++]=(gas[i][j].MarkTimeLeft);
        }
    }

    if(!(
        cudaPutToGpu(_this->_gpuData->Ro0,_this->_cpuBufer->Ro0,index)&&
        cudaPutToGpu(_this->_gpuData->P0,_this->_cpuBufer->P0,index)&&
        cudaPutToGpu(_this->_gpuData->Vx0,_this->_cpuBufer->Vx0,index)&&
        cudaPutToGpu(_this->_gpuData->Vy0,_this->_cpuBufer->Vy0,index)&&
        cudaPutToGpu(_this->_gpuData->Vs0,_this->_cpuBufer->Vs0,index)&&
        cudaPutToGpu(_this->_gpuData->Sc0,_this->_cpuBufer->Sc0,index)&&
        cudaPutToGpu(_this->_gpuData->Sc1,_this->_cpuBufer->Sc1,index)&&
        cudaPutToGpu(_this->_gpuData->Cs0,_this->_cpuBufer->Cs0,index)&&
        cudaPutToGpu(_this->_gpuData->Cs1,_this->_cpuBufer->Cs1,index)&&
        cudaPutToGpu(_this->_gpuData->R10,_this->_cpuBufer->R10,index)&&
        cudaPutToGpu(_this->_gpuData->R20,_this->_cpuBufer->R20,index)&&
        cudaPutToGpu(_this->_gpuData->R30,_this->_cpuBufer->R30,index)&&
        cudaPutToGpu(_this->_gpuData->G10,_this->_cpuBufer->G10,index)&&
        cudaPutToGpu(_this->_gpuData->G20,_this->_cpuBufer->G20,index)&&
        cudaPutToGpu(_this->_gpuData->G30,_this->_cpuBufer->G30,index)&&
        cudaPutToGpu(_this->_gpuData->B10,_this->_cpuBufer->B10,index)&&
        cudaPutToGpu(_this->_gpuData->B20,_this->_cpuBufer->B20,index)&&
        cudaPutToGpu(_this->_gpuData->B30,_this->_cpuBufer->B30,index)
        )
    )
    {
        logger(std::string("can't write data to gpu in device::operator="));
        *result = false;
    }

    pthread_exit(NULL);
}

bool Device::update()
{
    bool result;
    std::pair<Device *, bool *> input(this,&result);
    JOIN_THREAD(thread_update,&input,"Device::update")
    return result;
}

void * Device::thread_getGpuGas(void * arg)
{
    std::pair<Device *,std::vector<std::vector<Gas> > *> * input=(std::pair<Device *,std::vector<std::vector<Gas> > *> *) arg;

    Device * _this = input->first;
    std::vector<std::vector<Gas> > * gas = input->second;

    if(!cudaSelectGpu(_this->_gpuId))
    {
        logger(std::string("can't select gpu in Device::thread_getGpuGas on device "), &_this->_gpuId);
        pthread_exit(NULL);
    }



    if(!_this->_gpuData.get()||!_this->_cpuData)
    {
        logger(std::string("zero pointer gpuData or cpuData from device::getGpuGas"));
        pthread_exit(NULL);
    }

    int size = _this->_cpuData->getSize().x*(_this->getHeight()+1);

    if(!(
    cudaGetFromGpu(_this->_cpuBufer->Ro0,_this->_gpuData->Ro0,size)&&
    cudaGetFromGpu(_this->_cpuBufer->P0,_this->_gpuData->P0,size)&&
    cudaGetFromGpu(_this->_cpuBufer->Vx0,_this->_gpuData->Vx0,size)&&
    cudaGetFromGpu(_this->_cpuBufer->Vy0,_this->_gpuData->Vy0,size)&&
    cudaGetFromGpu(_this->_cpuBufer->Vs0,_this->_gpuData->Vs0,size)&&
    cudaGetFromGpu(_this->_cpuBufer->Sc0,_this->_gpuData->Sc0,size)&&
    cudaGetFromGpu(_this->_cpuBufer->Cs0,_this->_gpuData->Cs0,size)&&
    cudaGetFromGpu(_this->_cpuBufer->R10,_this->_gpuData->R10,size)&&
    cudaGetFromGpu(_this->_cpuBufer->R20,_this->_gpuData->R20,size)&&
    cudaGetFromGpu(_this->_cpuBufer->R30,_this->_gpuData->R30,size)&&
    cudaGetFromGpu(_this->_cpuBufer->G10,_this->_gpuData->G10,size)&&
    cudaGetFromGpu(_this->_cpuBufer->G20,_this->_gpuData->G20,size)&&
    cudaGetFromGpu(_this->_cpuBufer->G30,_this->_gpuData->G30,size)&&
    cudaGetFromGpu(_this->_cpuBufer->B10,_this->_gpuData->B10,size)&&
    cudaGetFromGpu(_this->_cpuBufer->B20,_this->_gpuData->B20,size)&&
    cudaGetFromGpu(_this->_cpuBufer->B30,_this->_gpuData->B30,size)&&
    cudaGetFromGpu(_this->_cpuBufer->Cs1,_this->_gpuData->Cs1,size)&&
    cudaGetFromGpu(_this->_cpuBufer->Sc1,_this->_gpuData->Sc1,size)))
    {
        logger(std::string("can't get gas from gpu in device::getGpuGas"));
        pthread_exit(NULL);
    }

    int index=_this->_cpuData->getSize().x;

    for(int j=_this->_stringMarks.x;j<_this->_stringMarks.y;++j)
    {
        for(int i=0;i<_this->_cpuData->getSize().x;++i)
        {
            (*gas)[i][j].Ro=_this->_cpuBufer->Ro0[index];
            (*gas)[i][j].P=_this->_cpuBufer->P0[index];
            (*gas)[i][j].Vx=_this->_cpuBufer->Vx0[index];
            (*gas)[i][j].Vy=_this->_cpuBufer->Vy0[index];
            (*gas)[i][j].Vs=_this->_cpuBufer->Vs0[index];
            (*gas)[i][j].Sc=_this->_cpuBufer->Sc0[index];
            (*gas)[i][j].Cs=_this->_cpuBufer->Cs0[index];
            (*gas)[i][j].R1=_this->_cpuBufer->R10[index];
            (*gas)[i][j].R2=_this->_cpuBufer->R20[index];
            (*gas)[i][j].R3=_this->_cpuBufer->R30[index];
            (*gas)[i][j].G1=_this->_cpuBufer->G10[index];
            (*gas)[i][j].G2=_this->_cpuBufer->G20[index];
            (*gas)[i][j].G3=_this->_cpuBufer->G30[index];
            (*gas)[i][j].B1=_this->_cpuBufer->B10[index];
            (*gas)[i][j].B2=_this->_cpuBufer->B20[index];
            (*gas)[i][j].B3=_this->_cpuBufer->B30[index];
            (*gas)[i][j].FullTimeLeft=_this->_cpuBufer->Cs1[index];
            (*gas)[i][j].MarkTimeLeft=_this->_cpuBufer->Sc1[index];
            (*gas)[i][j].check();
            ++index;
        }
    }

    pthread_exit(NULL);
}

void Device::getGpuGas(std::vector<std::vector<Gas> > & gas)
{
    std::pair<Device *,std::vector<std::vector<Gas> > *> input(this,&gas);
    JOIN_THREAD(thread_getGpuGas,&input,"Device::getGpuGas")
}

void * Device::thread_getColors(void * arg)
{
    std::pair<Device *,std::vector<std::vector<float> > *> * input=(std::pair<Device *,std::vector<std::vector<float> > *> *) arg;

    Device * _this = input->first;
    std::vector<std::vector<float> > * colors = input->second;

    if(!cudaSelectGpu(_this->_gpuId))
    {
        logger(std::string("can't select gpu in Device::thread_getColors on device "), &_this->_gpuId);
        pthread_exit(NULL);
    }

    if(!_this->_gpuData.get())
    {
        logger(std::string("zero pointer on gpuData from device::getColors"));
        pthread_exit(NULL);
    }

    int size = _this->_cpuData->getSize().x*_this->getHeight();

    //читаю все кроме последней строки
    if(!(
        cudaGetFromGpu(_this->_cpuBufer->R11,_this->_gpuData->R11,size+_this->_cpuData->getSize().x)&&
        cudaGetFromGpu(_this->_cpuBufer->G11,_this->_gpuData->G11,size+_this->_cpuData->getSize().x)&&
        cudaGetFromGpu(_this->_cpuBufer->B11,_this->_gpuData->B11,size+_this->_cpuData->getSize().x)))
    {
        logger(std::string("cant read colors from gpu in device::getColors"));
        pthread_exit(NULL);
    }

    int dest=_this->_stringMarks.x*_this->_cpuData->getSize().x;

    memcpy(&((*colors)[0][dest]),_this->_cpuBufer->R11+_this->_cpuData->getSize().x,size*sizeof(float));
    memcpy(&((*colors)[1][dest]),_this->_cpuBufer->G11+_this->_cpuData->getSize().x,size*sizeof(float));
    memcpy(&((*colors)[2][dest]),_this->_cpuBufer->B11+_this->_cpuData->getSize().x,size*sizeof(float));

    pthread_exit(NULL);
}

void Device::getColors(std::vector<std::vector<float> > & colors)
{
    std::pair<Device *,std::vector<std::vector<float> > *> input(this,&colors);
    JOIN_THREAD(thread_getColors,&input,"Device::getColors")
}

Device::Device()
{
    _upper=_downer=0;
    int error;
    SAFE_CALL(pthread_attr_init( &_attr ),"pthread_attr_init","Device::init")
    SAFE_CALL(pthread_attr_setdetachstate( &_attr, PTHREAD_CREATE_JOINABLE ),"pthread_attr_setdetachstate","Device::init")
}

Device::Device(const Device & in)
{
    _upper=_downer=0;
    int error;
    SAFE_CALL(pthread_attr_init( &_attr ),"pthread_attr_init","Device::init")
    SAFE_CALL(pthread_attr_setdetachstate( &_attr, PTHREAD_CREATE_JOINABLE ),"pthread_attr_setdetachstate","Device::init")
}

void * Device::thread_gpuFree(void * arg)
{
    Device * _this=(Device *)arg;

    if(!_this->_gpuData.get())
        pthread_exit(NULL);

    if(!cudaSelectGpu(_this->_gpuId))
    {
        logger(std::string("can't select gpu in Device::thread_gpuFree on device "), &_this->_gpuId);
        pthread_exit(NULL);
    }

    if(!(
    cudaFreeMemory(_this->_gpuData->Ro0)&&
    cudaFreeMemory(_this->_gpuData->Ro1)&&
    cudaFreeMemory(_this->_gpuData->P0)&&
    cudaFreeMemory(_this->_gpuData->P1)&&
    cudaFreeMemory(_this->_gpuData->Vx0)&&
    cudaFreeMemory(_this->_gpuData->Vx1)&&
    cudaFreeMemory(_this->_gpuData->Vy0)&&
    cudaFreeMemory(_this->_gpuData->Vy1)&&
    cudaFreeMemory(_this->_gpuData->Vs0)&&
    cudaFreeMemory(_this->_gpuData->Vs1)&&
    cudaFreeMemory(_this->_gpuData->Sc0)&&
    cudaFreeMemory(_this->_gpuData->Sc1)&&
    cudaFreeMemory(_this->_gpuData->Cs0)&&
    cudaFreeMemory(_this->_gpuData->Cs1)&&
    cudaFreeMemory(_this->_gpuData->R10)&&
    cudaFreeMemory(_this->_gpuData->R11)&&
    cudaFreeMemory(_this->_gpuData->R20)&&
    cudaFreeMemory(_this->_gpuData->R21)&&
    cudaFreeMemory(_this->_gpuData->R30)&&
    cudaFreeMemory(_this->_gpuData->R31)&&
    cudaFreeMemory(_this->_gpuData->G10)&&
    cudaFreeMemory(_this->_gpuData->G11)&&
    cudaFreeMemory(_this->_gpuData->G20)&&
    cudaFreeMemory(_this->_gpuData->G21)&&
    cudaFreeMemory(_this->_gpuData->G30)&&
    cudaFreeMemory(_this->_gpuData->G31)&&
    cudaFreeMemory(_this->_gpuData->B10)&&
    cudaFreeMemory(_this->_gpuData->B11)&&
    cudaFreeMemory(_this->_gpuData->B20)&&
    cudaFreeMemory(_this->_gpuData->B21)&&
    cudaFreeMemory(_this->_gpuData->B30)&&
    cudaFreeMemory(_this->_gpuData->B31)&&
    cudaFreeMemory(_this->_gpuData->bRo)&&
    cudaFreeMemory(_this->_gpuData->bVx)&&
    cudaFreeMemory(_this->_gpuData->bVy)&&
    cudaFreeMemory(_this->_gpuData->bAx)&&
    cudaFreeMemory(_this->_gpuData->bAy)&&

    cudaFreeHostMemory(_this->_cpuBufer->Ro0)&&
    cudaFreeHostMemory(_this->_cpuBufer->P0)&&
    cudaFreeHostMemory(_this->_cpuBufer->Vx0)&&
    cudaFreeHostMemory(_this->_cpuBufer->Vy0)&&
    cudaFreeHostMemory(_this->_cpuBufer->Vs0)&&
    cudaFreeHostMemory(_this->_cpuBufer->Sc0)&&
    cudaFreeHostMemory(_this->_cpuBufer->Cs0)&&
    cudaFreeHostMemory(_this->_cpuBufer->R10)&&
    cudaFreeHostMemory(_this->_cpuBufer->R20)&&
    cudaFreeHostMemory(_this->_cpuBufer->R30)&&
    cudaFreeHostMemory(_this->_cpuBufer->G10)&&
    cudaFreeHostMemory(_this->_cpuBufer->G20)&&
    cudaFreeHostMemory(_this->_cpuBufer->G30)&&
    cudaFreeHostMemory(_this->_cpuBufer->B10)&&
    cudaFreeHostMemory(_this->_cpuBufer->B20)&&
    cudaFreeHostMemory(_this->_cpuBufer->B30)&&
    cudaFreeHostMemory(_this->_cpuBufer->Ro1)&&
    cudaFreeHostMemory(_this->_cpuBufer->P1)&&
    cudaFreeHostMemory(_this->_cpuBufer->Vx1)&&
    cudaFreeHostMemory(_this->_cpuBufer->Vy1)&&
    cudaFreeHostMemory(_this->_cpuBufer->Vs1)&&
    cudaFreeHostMemory(_this->_cpuBufer->Sc1)&&
    cudaFreeHostMemory(_this->_cpuBufer->Cs1)&&
    cudaFreeHostMemory(_this->_cpuBufer->R11)&&
    cudaFreeHostMemory(_this->_cpuBufer->R21)&&
    cudaFreeHostMemory(_this->_cpuBufer->R31)&&
    cudaFreeHostMemory(_this->_cpuBufer->G11)&&
    cudaFreeHostMemory(_this->_cpuBufer->G21)&&
    cudaFreeHostMemory(_this->_cpuBufer->G31)&&
    cudaFreeHostMemory(_this->_cpuBufer->B11)&&
    cudaFreeHostMemory(_this->_cpuBufer->B21)&&
    cudaFreeHostMemory(_this->_cpuBufer->B31)&&
    cudaFreeHostMemory(_this->_cpuBufer->bRo)&&
    cudaFreeHostMemory(_this->_cpuBufer->bVx)&&
    cudaFreeHostMemory(_this->_cpuBufer->bVy)&&
    cudaFreeHostMemory(_this->_cpuBufer->bAx)&&
    cudaFreeHostMemory(_this->_cpuBufer->bAy)&&
    cudaFreeHostMemory(_this->_cpuBufer->gravity)&&

    cudaFreeHostMemory(_this->_upMark->Ro0)&&
    cudaFreeHostMemory(_this->_upMark->P0)&&
    cudaFreeHostMemory(_this->_upMark->Vx0)&&
    cudaFreeHostMemory(_this->_upMark->Vy0)&&
    cudaFreeHostMemory(_this->_upMark->Vs0)&&
    cudaFreeHostMemory(_this->_upMark->Sc0)&&
    cudaFreeHostMemory(_this->_upMark->Cs0)&&
    cudaFreeHostMemory(_this->_upMark->R10)&&
    cudaFreeHostMemory(_this->_upMark->R20)&&
    cudaFreeHostMemory(_this->_upMark->R30)&&
    cudaFreeHostMemory(_this->_upMark->G10)&&
    cudaFreeHostMemory(_this->_upMark->G20)&&
    cudaFreeHostMemory(_this->_upMark->G30)&&
    cudaFreeHostMemory(_this->_upMark->B10)&&
    cudaFreeHostMemory(_this->_upMark->B20)&&
    cudaFreeHostMemory(_this->_upMark->B30)&&
    cudaFreeHostMemory(_this->_upMark->Ro1)&&
    cudaFreeHostMemory(_this->_upMark->P1)&&
    cudaFreeHostMemory(_this->_upMark->Vx1)&&
    cudaFreeHostMemory(_this->_upMark->Vy1)&&
    cudaFreeHostMemory(_this->_upMark->Vs1)&&
    cudaFreeHostMemory(_this->_upMark->Sc1)&&
    cudaFreeHostMemory(_this->_upMark->Cs1)&&
    cudaFreeHostMemory(_this->_upMark->R11)&&
    cudaFreeHostMemory(_this->_upMark->R21)&&
    cudaFreeHostMemory(_this->_upMark->R31)&&
    cudaFreeHostMemory(_this->_upMark->G11)&&
    cudaFreeHostMemory(_this->_upMark->G21)&&
    cudaFreeHostMemory(_this->_upMark->G31)&&
    cudaFreeHostMemory(_this->_upMark->B11)&&
    cudaFreeHostMemory(_this->_upMark->B21)&&
    cudaFreeHostMemory(_this->_upMark->B31)&&

    cudaFreeHostMemory(_this->_dwMark->Ro0)&&
    cudaFreeHostMemory(_this->_dwMark->P0)&&
    cudaFreeHostMemory(_this->_dwMark->Vx0)&&
    cudaFreeHostMemory(_this->_dwMark->Vy0)&&
    cudaFreeHostMemory(_this->_dwMark->Vs0)&&
    cudaFreeHostMemory(_this->_dwMark->Sc0)&&
    cudaFreeHostMemory(_this->_dwMark->Cs0)&&
    cudaFreeHostMemory(_this->_dwMark->R10)&&
    cudaFreeHostMemory(_this->_dwMark->R20)&&
    cudaFreeHostMemory(_this->_dwMark->R30)&&
    cudaFreeHostMemory(_this->_dwMark->G10)&&
    cudaFreeHostMemory(_this->_dwMark->G20)&&
    cudaFreeHostMemory(_this->_dwMark->G30)&&
    cudaFreeHostMemory(_this->_dwMark->B10)&&
    cudaFreeHostMemory(_this->_dwMark->B20)&&
    cudaFreeHostMemory(_this->_dwMark->B30)&&
    cudaFreeHostMemory(_this->_dwMark->Ro1)&&
    cudaFreeHostMemory(_this->_dwMark->P1)&&
    cudaFreeHostMemory(_this->_dwMark->Vx1)&&
    cudaFreeHostMemory(_this->_dwMark->Vy1)&&
    cudaFreeHostMemory(_this->_dwMark->Vs1)&&
    cudaFreeHostMemory(_this->_dwMark->Sc1)&&
    cudaFreeHostMemory(_this->_dwMark->Cs1)&&
    cudaFreeHostMemory(_this->_dwMark->R11)&&
    cudaFreeHostMemory(_this->_dwMark->R21)&&
    cudaFreeHostMemory(_this->_dwMark->R31)&&
    cudaFreeHostMemory(_this->_dwMark->G11)&&
    cudaFreeHostMemory(_this->_dwMark->G21)&&
    cudaFreeHostMemory(_this->_dwMark->G31)&&
    cudaFreeHostMemory(_this->_dwMark->B11)&&
    cudaFreeHostMemory(_this->_dwMark->B21)&&
    cudaFreeHostMemory(_this->_dwMark->B31)))
    {
        logger(std::string("can't free memory on gpu in Device::thread_gpuFree on device "), &_this->_gpuId);
        pthread_exit(NULL);
    }


    pthread_exit(NULL);
}

Device::~Device()
{
    JOIN_THREAD(thread_gpuFree,this,"Device::~Device")
    SAFE_CALL(pthread_attr_destroy(&_attr),"pthread_attr_destroy","Device::~Device")
}

//#define DEBUG

void * Device::thread_stepTo(void * arg)
{
    Device * _this=(Device *)arg;

    if(!cudaSelectGpu(_this->_gpuId))
    {
        logger(std::string("can't select gpu in Device::thread_stepTo on device "), &_this->_gpuId);
        pthread_exit(NULL);
    }

    int taskSize=_this->_cpuData->getSize().x*_this->getHeight();
    int width=_this->_cpuData->getSize().x;

#ifdef DEBUG
    TestStory debug(Position(width,_this->getHeight()+2), Position(0,0), 1, 1, width*(_this->getHeight()+2), width,true);
    int xstep=0;
    int ystep=0;
#endif

    while(!_this->_cpuData->isFrame())
    {
        float dt=_this->_cpuData->getDt();

        for(int direction=0;direction<2;++direction)
        {

            if(_this->_transponate)
            {
                _this->getMarks();
                //прочитал свои границы
                pthread_barrier_wait(&BARRIER);/////////////////////////////
                _this->setMarks(_this->_upper?_this->_upper->getDownMark():0, _this->_downer?_this->_downer->getUpMark():0);
                //установил граничные условия от соседей
            }

#ifdef DEBUG
    debug.dy=!_this->_transponate;
    debug.a_first(_this->_gpuData.get());
#endif

            cudaGasCorrect(_this->_gpuData->getGpuGasCorrectData(_this->_transponate,width,dt),taskSize);
            //посчитал коррекцию

            if(_this->_transponate)
            {
                _this->getGasCorrectedMarks();
                //прочитал свои граничные откорректированные значения газа
                pthread_barrier_wait(&BARRIER);/////////////////////////////
                _this->setGasCorrectedMarks(_this->_upper?_this->_upper->getDownMark():0, _this->_downer?_this->_downer->getUpMark():0);
                //установил граничные откорректированные значения газа от соседей
            }

#ifdef DEBUG
    debug.b_gasCorrect(_this->_gpuData.get());
#endif

            cudaGasRieman(_this->_gpuData->getGpuGasRiemanData(_this->_transponate,width),taskSize);
            //расчитал значения между ячейками

            if(_this->_transponate)
            {
                GpuGasRiemanData markRieman=_this->_gpuData->getGpuGasRiemanData(_this->_transponate,width);
                markRieman.shift=taskSize;
                cudaGasRieman(markRieman,width);
                //расчитал значения между ячейками на нижней границе
            }

#ifdef DEBUG
    debug.c_gasRieman(_this->_gpuData.get());
#endif

            cudaGasSolver(_this->_gpuData->getGpuGasSolverData(_this->_transponate,width,dt,_this->_scenario->getMaxVisc()),taskSize);
            //рассчитал значения газа в следующее время

            _this->getGasSolvedMarks();
            //прочитал свои граничные новые значения газа
            pthread_barrier_wait(&BARRIER);/////////////////////////////
            _this->setGasSolvedMarks(_this->_upper?_this->_upper->getDownMark():0, _this->_downer?_this->_downer->getUpMark():0);
            //установил граничные новые значения газа от соседей

#ifdef DEBUG
    debug.e_gasSolve(_this->_gpuData.get());
#endif

            cudaViscBuilder(_this->_gpuData->getGpuGasViscData(_this->_transponate,width,dt),taskSize);
            //рассчитал воздействие вязкости

#ifdef DEBUG
debug.d_gasVisc(_this->_gpuData.get());
#endif

            cudaDataCorrect(_this->_gpuData->getGpuDataCorrectData(_this->_transponate,width,_this->_scenario->getParamsSchemeCorrection()),taskSize);
            //рассчитал откорректированные значения данных

            if(_this->_transponate)
            {
                _this->getDataCorrectedMarks();
                //прочитал свои откорректированные данные
                pthread_barrier_wait(&BARRIER);/////////////////////////////
                _this->setDataCorrectedMarks(_this->_upper?_this->_upper->getDownMark():0, _this->_downer?_this->_downer->getUpMark():0);
                //установил граничные откорректированные значения данных от соседей
            }

#ifdef DEBUG
    debug.f_dataCorrect(_this->_gpuData.get());
#endif

            cudaDataSolver(_this->_gpuData->getGpuDataSolverData(_this->_transponate,width,dt),taskSize);
            //рассчитал значения данных в будущем

#ifdef DEBUG
    debug.g_dataSolve(_this->_gpuData.get());
    if(_this->_transponate==false)
    {
        xstep++;
    }
    else
    {
        ystep++;
    }
#endif

            _this->_transponate = direction ? _this->_transponate : !_this->_transponate;
            //изменил направление рассчета

        }

        if(_this->_cpuData->getBodyCount()>0)
        {
            _this->_cpuData->makeConstBody(_this->_cpuBufer->gravity);
            cudaSetConstData(_this->_cpuBufer->gravity);
            cudaGravity(_this->_gpuData->getGpuGravityData(_this->_stringMarks.x,_this->_cpuData->getBodyCount(),width,dt),taskSize);
            cudaBodyGravity(_this->_gpuData->getGpuBodyGravityData(_this->_stringMarks.x,_this->_stringMarks.y,_this->_cpuData->getBodyCount(),dt),4096);
            cudaGetFromGpu(_this->_cpuBufer->bRo,_this->_gpuData->bRo,4096);
            cudaGetFromGpu(_this->_cpuBufer->bVx,_this->_gpuData->bVx,4096);
            cudaGetFromGpu(_this->_cpuBufer->bVy,_this->_gpuData->bVy,4096);
            cudaGetFromGpu(_this->_cpuBufer->bAx,_this->_gpuData->bAx,4096);
            cudaGetFromGpu(_this->_cpuBufer->bAy,_this->_gpuData->bAy,4096);
            for(int b=0;b<_this->_cpuData->getBodyCount();++b)
            {
                CpuGravityBody * body=_this->_cpuData->getBody(b);
                if(body->y>=_this->_stringMarks.x&&body->y<_this->_stringMarks.y)
                {
                    if(body->x>=0&&body->x<width)
                    {
                        body->gasRo=_this->_cpuBufer->bRo[b];
                        body->gasVx=_this->_cpuBufer->bVx[b];
                        body->gasVy=_this->_cpuBufer->bVy[b];
                    }
                    body->vx+=_this->_cpuBufer->bAx[b]*body->bodyV;
                    body->vy+=_this->_cpuBufer->bAy[b]*body->bodyV;
                    body->x+=body->getVx()*_this->_cpuData->getDt();
                    body->y+=body->getVy()*_this->_cpuData->getDt();
                }
            }
        }

        cudaFinalize(_this->_gpuData->getGpuFinalizeData(_this->_transponate,width,dt),taskSize);

#ifdef DEBUG
    debug.h_finalize(_this->_gpuData.get());
#endif

        //шаг твердого тела и расчет dt
        int blockSize=1;
        for(unsigned int i=taskSize;i>1;i/=blockSize)
        {
            blockSize=cudaDtIteration(_this->_gpuData->getGpuDtData(width),i);
            cudaDtCopy(_this->_gpuData->getGpuDtData(width),i);
            //поиск минимального dt
        }
        _this->_cpuData->setNextDt(cudaGetMinDt(_this->_gpuData->getGpuDtData(width).data,_this->_cpuBufer->getGpuDtData(width).data,width));
        //получил минимальное dt

        pthread_barrier_wait(&BARRIER);/////////////////////////////

        _this->_cpuData->setDt();
    }
    cudaRender(_this->_gpuData->getGpuRenderData(_this->_transponate,width),taskSize);
    pthread_exit(NULL);
}

bool Device::step()
{
    int error;
    SAFE_CALL(pthread_create(&_thread, &_attr, thread_stepTo, (void *)this),"pthread_create","Device::stepTo")
    return !error;
}

void Device::wait()
{
    int error;
    SAFE_CALL(pthread_join(_thread, NULL),"pthread_join","Device::wait")       
}

void * thread_getGpusInfo(void * arg)
{
    std::vector<std::string> * out = (std::vector<std::string> *)arg;
    try
    {
        *out=cudaGetGpusInfo();
    }
    catch(...)
    {
        logger("cant take gpus info in thread_getGpusInfo");
        throw;
    }

    pthread_exit(NULL);
}

std::vector<std::string> Device::getGpusInfo()
{
    std::vector<std::string> out;

    JOIN_THREAD(thread_getGpusInfo,&out,"Device::getGpusInfo")

    return out;
}

void * thread_getGpuInfo(void * arg)
{
    GpuInfo * gpuInfo = (GpuInfo *)arg;
    if(!cudaSelectGpu(gpuInfo->gpuId))
    {
        logger(std::string("can't select gpu in thread_getGpuInfo on device "), &gpuInfo->gpuId);
        pthread_exit(NULL);
    }

    try
    {
        *gpuInfo=cudaGetGpuInfo(gpuInfo->gpuId);
    }
    catch(...)
    {
        logger("cant take gpu info in thread_getGpuInfo");
        throw;
    }

    pthread_exit(NULL);
}

GpuInfo Device::getGpuInfo(int id)
{
    GpuInfo out;
    out.gpuId=id;

    JOIN_THREAD(thread_getGpuInfo,&out,"Device::getGpuInfo")

    return out;
}

bool Device::setDeviceCount(int count)
{
    if(count>0)
    {
        int error = pthread_barrier_init(&BARRIER, NULL, count);
        if(!error)
            return true;
        else
            std::cout<<error<<" error in Device::setDeviceCount"<<std::endl;
    }
    else
    {
        logger("cant init barrier with count < 1 in Device::setDeviceCount");
    }
    return false;
}

void Device::getMarks()
{
    int len=0;
    if(_downer)
    {//копирую нижнюю строку в первую на gpu, если внизу кто-то есть
        GpuCopyStringData data=_gpuData->getGpuCopyStringDataBase(0,getHeight());
        cudaCopyStrings(data,_cpuData->getSize().x);
    }

    if(_downer||_upper)
    {// читаю первые две строки из гпу
        len=_cpuData->getSize().x*2;

        cudaGetFromGpu(_cpuBufer->Ro0,_gpuData->Ro0,len);
        cudaGetFromGpu(_cpuBufer->P0,_gpuData->P0,len);
        cudaGetFromGpu(_cpuBufer->Vx0,_gpuData->Vx0,len);
        cudaGetFromGpu(_cpuBufer->Vy0,_gpuData->Vy0,len);
        cudaGetFromGpu(_cpuBufer->Vs0,_gpuData->Vs0,len);
        cudaGetFromGpu(_cpuBufer->Sc0,_gpuData->Sc0,len);
        cudaGetFromGpu(_cpuBufer->Cs0,_gpuData->Cs0,len);
        cudaGetFromGpu(_cpuBufer->R10,_gpuData->R10,len);
        cudaGetFromGpu(_cpuBufer->R20,_gpuData->R20,len);
        cudaGetFromGpu(_cpuBufer->R30,_gpuData->R30,len);
        cudaGetFromGpu(_cpuBufer->G10,_gpuData->G10,len);
        cudaGetFromGpu(_cpuBufer->G20,_gpuData->G20,len);
        cudaGetFromGpu(_cpuBufer->G30,_gpuData->G30,len);
        cudaGetFromGpu(_cpuBufer->B10,_gpuData->B10,len);
        cudaGetFromGpu(_cpuBufer->B20,_gpuData->B20,len);
        cudaGetFromGpu(_cpuBufer->B30,_gpuData->B30,len);
    }

    if(_downer)
    {//если есть нижний сосед то заполняю нижнюю границу
        len=_cpuData->getSize().x*sizeof(float);

        memcpy(_dwMark->Ro0,_cpuBufer->Ro0,len);
        memcpy(_dwMark->P0,_cpuBufer->P0,len);
        memcpy(_dwMark->Vx0,_cpuBufer->Vx0,len);
        memcpy(_dwMark->Vy0,_cpuBufer->Vy0,len);
        memcpy(_dwMark->Vs0,_cpuBufer->Vs0,len);
        memcpy(_dwMark->Sc0,_cpuBufer->Sc0,len);
        memcpy(_dwMark->Cs0,_cpuBufer->Cs0,len);
        memcpy(_dwMark->R10,_cpuBufer->R10,len);
        memcpy(_dwMark->R20,_cpuBufer->R20,len);
        memcpy(_dwMark->R30,_cpuBufer->R30,len);
        memcpy(_dwMark->G10,_cpuBufer->G10,len);
        memcpy(_dwMark->G20,_cpuBufer->G20,len);
        memcpy(_dwMark->G30,_cpuBufer->G30,len);
        memcpy(_dwMark->B10,_cpuBufer->B10,len);
        memcpy(_dwMark->B20,_cpuBufer->B20,len);
        memcpy(_dwMark->B30,_cpuBufer->B30,len);
    }

    if(_upper)
    {//если есть верхний сосед то заполняю верхнюю границу
        len=_cpuData->getSize().x*sizeof(float);

        memcpy(_upMark->Ro0,_cpuBufer->Ro0+_cpuData->getSize().x,len);
        memcpy(_upMark->P0,_cpuBufer->P0+_cpuData->getSize().x,len);
        memcpy(_upMark->Vx0,_cpuBufer->Vx0+_cpuData->getSize().x,len);
        memcpy(_upMark->Vy0,_cpuBufer->Vy0+_cpuData->getSize().x,len);
        memcpy(_upMark->Vs0,_cpuBufer->Vs0+_cpuData->getSize().x,len);
        memcpy(_upMark->Sc0,_cpuBufer->Sc0+_cpuData->getSize().x,len);
        memcpy(_upMark->Cs0,_cpuBufer->Cs0+_cpuData->getSize().x,len);
        memcpy(_upMark->R10,_cpuBufer->R10+_cpuData->getSize().x,len);
        memcpy(_upMark->R20,_cpuBufer->R20+_cpuData->getSize().x,len);
        memcpy(_upMark->R30,_cpuBufer->R30+_cpuData->getSize().x,len);
        memcpy(_upMark->G10,_cpuBufer->G10+_cpuData->getSize().x,len);
        memcpy(_upMark->G20,_cpuBufer->G20+_cpuData->getSize().x,len);
        memcpy(_upMark->G30,_cpuBufer->G30+_cpuData->getSize().x,len);
        memcpy(_upMark->B10,_cpuBufer->B10+_cpuData->getSize().x,len);
        memcpy(_upMark->B20,_cpuBufer->B20+_cpuData->getSize().x,len);
        memcpy(_upMark->B30,_cpuBufer->B30+_cpuData->getSize().x,len);
    }
}

void Device::getGasCorrectedMarks()
{
    int len=0;
    if(_downer)
    {//копирую нижнюю строку в первую на gpu, если внизу кто-то есть
        GpuCopyStringData data=_gpuData->getGpuCopyStringDataCorrectGas(0,getHeight());
        cudaCopyStrings(data,_cpuData->getSize().x);
    }

    if(_downer||_upper)
    {// читаю первые две строки из гпу
        len=_cpuData->getSize().x*2;
            cudaGetFromGpu(_cpuBufer->Ro1,_gpuData->Ro1,len);
            cudaGetFromGpu(_cpuBufer->P1,_gpuData->P1,len);
            cudaGetFromGpu(_cpuBufer->Vx1,_gpuData->Vx1,len);
            cudaGetFromGpu(_cpuBufer->Vy1,_gpuData->Vy1,len);
            cudaGetFromGpu(_cpuBufer->Vs1,_gpuData->Vs1,len);
    }

    if(_downer)
    {//если есть нижний сосед то заполняю нижнюю границу
        len=_cpuData->getSize().x*sizeof(float);

        memcpy(_dwMark->Ro1,_cpuBufer->Ro1,len);
        memcpy(_dwMark->P1,_cpuBufer->P1,len);
        memcpy(_dwMark->Vx1,_cpuBufer->Vx1,len);
        memcpy(_dwMark->Vy1,_cpuBufer->Vy1,len);
        memcpy(_dwMark->Vs1,_cpuBufer->Vs1,len);
    }

    if(_upper)
    {//если есть верхний сосед то заполняю верхнюю границу
        len=_cpuData->getSize().x*sizeof(float);

        memcpy(_upMark->Ro1,_cpuBufer->Ro1+_cpuData->getSize().x,len);
        memcpy(_upMark->P1,_cpuBufer->P1+_cpuData->getSize().x,len);
        memcpy(_upMark->Vx1,_cpuBufer->Vx1+_cpuData->getSize().x,len);
        memcpy(_upMark->Vy1,_cpuBufer->Vy1+_cpuData->getSize().x,len);
        memcpy(_upMark->Vs1,_cpuBufer->Vs1+_cpuData->getSize().x,len);
    }
}


void Device::getGasSolvedMarks()
{
    int len=0;
    if(_downer)
    {//копирую нижнюю строку в первую на gpu, если внизу кто-то есть
        GpuCopyStringData data=_gpuData->getGpuCopyStringDataSolvedGas(0,getHeight());
        cudaCopyStrings(data,_cpuData->getSize().x);
    }

    if(_downer||_upper)
    {// читаю первые две строки из гпу
        len=_cpuData->getSize().x*2;

        cudaGetFromGpu(_cpuBufer->Vx0,_gpuData->Vx0,len);
        cudaGetFromGpu(_cpuBufer->Vy0,_gpuData->Vy0,len);
        cudaGetFromGpu(_cpuBufer->Vs0,_gpuData->Vs0,len);
    }

    if(_downer)
    {//если есть нижний сосед то заполняю нижнюю границу
        len=_cpuData->getSize().x*sizeof(float);
        memcpy(_dwMark->Vs0,_cpuBufer->Vs0,len);
        memcpy(_dwMark->Vx0,_cpuBufer->Vx0,len);
        memcpy(_dwMark->Vy0,_cpuBufer->Vy0,len);
    }

    if(_upper)
    {//если есть верхний сосед то заполняю верхнюю границу
        len=_cpuData->getSize().x*sizeof(float);
        memcpy(_upMark->Vs0,_cpuBufer->Vs0+_cpuData->getSize().x,len);
        memcpy(_upMark->Vx0,_cpuBufer->Vx0+_cpuData->getSize().x,len);
        memcpy(_upMark->Vy0,_cpuBufer->Vy0+_cpuData->getSize().x,len);
    }
}

void Device::getDataCorrectedMarks()
{
    int len=0;
    if(_downer)
    {//копирую нижнюю строку в первую на gpu, если внизу кто-то есть
        GpuCopyStringData data=_gpuData->getGpuCopyStringDataCorrectData(0,getHeight());
        cudaCopyStrings(data,_cpuData->getSize().x);
    }

    if(_downer||_upper)
    {// читаю первые две строки из гпу
        len=_cpuData->getSize().x*2;

        cudaGetFromGpu(_cpuBufer->Vs1,_gpuData->Vs1,len);
        cudaGetFromGpu(_cpuBufer->Sc1,_gpuData->Sc1,len);
        cudaGetFromGpu(_cpuBufer->Cs1,_gpuData->Cs1,len);
        cudaGetFromGpu(_cpuBufer->R11,_gpuData->R11,len);
        cudaGetFromGpu(_cpuBufer->R21,_gpuData->R21,len);
        cudaGetFromGpu(_cpuBufer->R31,_gpuData->R31,len);
        cudaGetFromGpu(_cpuBufer->G11,_gpuData->G11,len);
        cudaGetFromGpu(_cpuBufer->G21,_gpuData->G21,len);
        cudaGetFromGpu(_cpuBufer->G31,_gpuData->G31,len);
        cudaGetFromGpu(_cpuBufer->B11,_gpuData->B11,len);
        cudaGetFromGpu(_cpuBufer->B21,_gpuData->B21,len);
        cudaGetFromGpu(_cpuBufer->B31,_gpuData->B31,len);
    }

    if(_downer)
    {//если есть нижний сосед то заполняю нижнюю границу
        len=_cpuData->getSize().x*sizeof(float);

        memcpy(_dwMark->Vs1,_cpuBufer->Vs1,len);
        memcpy(_dwMark->Sc1,_cpuBufer->Sc1,len);
        memcpy(_dwMark->Cs1,_cpuBufer->Cs1,len);
        memcpy(_dwMark->R11,_cpuBufer->R11,len);
        memcpy(_dwMark->R21,_cpuBufer->R21,len);
        memcpy(_dwMark->R31,_cpuBufer->R31,len);
        memcpy(_dwMark->G11,_cpuBufer->G11,len);
        memcpy(_dwMark->G21,_cpuBufer->G21,len);
        memcpy(_dwMark->G31,_cpuBufer->G31,len);
        memcpy(_dwMark->B11,_cpuBufer->B11,len);
        memcpy(_dwMark->B21,_cpuBufer->B21,len);
        memcpy(_dwMark->B31,_cpuBufer->B31,len);
    }

    if(_upper)
    {//если есть верхний сосед то заполняю верхнюю границу
        len=_cpuData->getSize().x*sizeof(float);

        memcpy(_upMark->Vs1,_cpuBufer->Vs1+_cpuData->getSize().x,len);
        memcpy(_upMark->Sc1,_cpuBufer->Sc1+_cpuData->getSize().x,len);
        memcpy(_upMark->Cs1,_cpuBufer->Cs1+_cpuData->getSize().x,len);
        memcpy(_upMark->R11,_cpuBufer->R11+_cpuData->getSize().x,len);
        memcpy(_upMark->R21,_cpuBufer->R21+_cpuData->getSize().x,len);
        memcpy(_upMark->R31,_cpuBufer->R31+_cpuData->getSize().x,len);
        memcpy(_upMark->G11,_cpuBufer->G11+_cpuData->getSize().x,len);
        memcpy(_upMark->G21,_cpuBufer->G21+_cpuData->getSize().x,len);
        memcpy(_upMark->G31,_cpuBufer->G31+_cpuData->getSize().x,len);
        memcpy(_upMark->B11,_cpuBufer->B11+_cpuData->getSize().x,len);
        memcpy(_upMark->B21,_cpuBufer->B21+_cpuData->getSize().x,len);
        memcpy(_upMark->B31,_cpuBufer->B31+_cpuData->getSize().x,len);
    }
}

void Device::setMarks(std::auto_ptr<GpuData> * up, std::auto_ptr<GpuData> * dw)
{
    int len=_cpuData->getSize().x;
    GpuCopyStringData data=_gpuData->getGpuCopyStringDataBase(getHeight()+1,getHeight());

    if(dw)
    {
        data.srcString=0;
        cudaPutToGpu(_gpuData->Ro0,(*dw)->Ro0,len);
        cudaPutToGpu(_gpuData->P0,(*dw)->P0,len);
        cudaPutToGpu(_gpuData->Vx0,(*dw)->Vx0,len);
        cudaPutToGpu(_gpuData->Vy0,(*dw)->Vy0,len);
        cudaPutToGpu(_gpuData->Vs0,(*dw)->Vs0,len);
        cudaPutToGpu(_gpuData->Sc0,(*dw)->Sc0,len);
        cudaPutToGpu(_gpuData->Cs0,(*dw)->Cs0,len);
        cudaPutToGpu(_gpuData->R10,(*dw)->R10,len);
        cudaPutToGpu(_gpuData->R20,(*dw)->R20,len);
        cudaPutToGpu(_gpuData->R30,(*dw)->R30,len);
        cudaPutToGpu(_gpuData->G10,(*dw)->G10,len);
        cudaPutToGpu(_gpuData->G20,(*dw)->G20,len);
        cudaPutToGpu(_gpuData->G30,(*dw)->G30,len);
        cudaPutToGpu(_gpuData->B10,(*dw)->B10,len);
        cudaPutToGpu(_gpuData->B20,(*dw)->B20,len);
        cudaPutToGpu(_gpuData->B30,(*dw)->B30,len);
    }

    cudaCopyStrings(data,_cpuData->getSize().x);


    ////////////////////////////////////////////////////////////

    len=_cpuData->getSize().x;

    if(up)
    {
        cudaPutToGpu(_gpuData->Ro0,(*up)->Ro0,len);
        cudaPutToGpu(_gpuData->P0,(*up)->P0,len);
        cudaPutToGpu(_gpuData->Vx0,(*up)->Vx0,len);
        cudaPutToGpu(_gpuData->Vy0,(*up)->Vy0,len);
        cudaPutToGpu(_gpuData->Vs0,(*up)->Vs0,len);
        cudaPutToGpu(_gpuData->Sc0,(*up)->Sc0,len);
        cudaPutToGpu(_gpuData->Cs0,(*up)->Cs0,len);
        cudaPutToGpu(_gpuData->R10,(*up)->R10,len);
        cudaPutToGpu(_gpuData->R20,(*up)->R20,len);
        cudaPutToGpu(_gpuData->R30,(*up)->R30,len);
        cudaPutToGpu(_gpuData->G10,(*up)->G10,len);
        cudaPutToGpu(_gpuData->G20,(*up)->G20,len);
        cudaPutToGpu(_gpuData->G30,(*up)->G30,len);
        cudaPutToGpu(_gpuData->B10,(*up)->B10,len);
        cudaPutToGpu(_gpuData->B20,(*up)->B20,len);
        cudaPutToGpu(_gpuData->B30,(*up)->B30,len);
    }
    else
    {
        data.srcString=1;
        data.dstString=0;
        cudaCopyStrings(data,_cpuData->getSize().x);
    }
}

void Device::setGasCorrectedMarks(std::auto_ptr<GpuData> * up, std::auto_ptr<GpuData> * dw)
{
    int len=_cpuData->getSize().x;
    GpuCopyStringData data=_gpuData->getGpuCopyStringDataCorrectGas(getHeight()+1,getHeight());

    if(dw)
    {
        data.srcString=0;

        cudaPutToGpu(_gpuData->Ro1,(*dw)->Ro1,len);
        cudaPutToGpu(_gpuData->P1,(*dw)->P1,len);
        cudaPutToGpu(_gpuData->Vx1,(*dw)->Vx1,len);
        cudaPutToGpu(_gpuData->Vy1,(*dw)->Vy1,len);
        cudaPutToGpu(_gpuData->Vs1,(*dw)->Vs1,len);
    }

    cudaCopyStrings(data,_cpuData->getSize().x);


    ////////////////////////////////////////////////////////////

    len=_cpuData->getSize().x;

    if(up)
    {
        cudaPutToGpu(_gpuData->Ro1,(*up)->Ro1,len);
        cudaPutToGpu(_gpuData->P1,(*up)->P1,len);
        cudaPutToGpu(_gpuData->Vx1,(*up)->Vx1,len);
        cudaPutToGpu(_gpuData->Vy1,(*up)->Vy1,len);
        cudaPutToGpu(_gpuData->Vs1,(*up)->Vs1,len);
    }
    else
    {
        data.srcString=1;
        data.dstString=0;
        cudaCopyStrings(data,_cpuData->getSize().x);
    }
}

void Device::setGasSolvedMarks(std::auto_ptr<GpuData> * up, std::auto_ptr<GpuData> * dw)
{
    int len=_cpuData->getSize().x;
    GpuCopyStringData data=_gpuData->getGpuCopyStringDataSolvedGas(getHeight()+1,getHeight());

    if(dw)
    {
        data.srcString=0;
        cudaPutToGpu(_gpuData->Vs0,(*dw)->Vs0,len);
        cudaPutToGpu(_gpuData->Vx0,(*dw)->Vx0,len);
        cudaPutToGpu(_gpuData->Vy0,(*dw)->Vy0,len);
    }

    cudaCopyStrings(data,_cpuData->getSize().x);


    ////////////////////////////////////////////////////////////

    len=_cpuData->getSize().x;

    if(up)
    {
        cudaPutToGpu(_gpuData->Vs0,(*up)->Vs0,len);
        cudaPutToGpu(_gpuData->Vx0,(*up)->Vx0,len);
        cudaPutToGpu(_gpuData->Vy0,(*up)->Vy0,len);
    }
    else
    {
        data.srcString=1;
        data.dstString=0;
        cudaCopyStrings(data,_cpuData->getSize().x);
    }
}

void Device::setDataCorrectedMarks(std::auto_ptr<GpuData> * up, std::auto_ptr<GpuData> * dw)
{
    int len=_cpuData->getSize().x;
    GpuCopyStringData data=_gpuData->getGpuCopyStringDataCorrectData(getHeight()+1,getHeight());

    if(dw)
    {
        data.srcString=0;
        cudaPutToGpu(_gpuData->Vs1,(*dw)->Vs1,len);
        cudaPutToGpu(_gpuData->Sc1,(*dw)->Sc1,len);
        cudaPutToGpu(_gpuData->Cs1,(*dw)->Cs1,len);
        cudaPutToGpu(_gpuData->R11,(*dw)->R11,len);
        cudaPutToGpu(_gpuData->R21,(*dw)->R21,len);
        cudaPutToGpu(_gpuData->R31,(*dw)->R31,len);
        cudaPutToGpu(_gpuData->G11,(*dw)->G11,len);
        cudaPutToGpu(_gpuData->G21,(*dw)->G21,len);
        cudaPutToGpu(_gpuData->G31,(*dw)->G31,len);
        cudaPutToGpu(_gpuData->B11,(*dw)->B11,len);
        cudaPutToGpu(_gpuData->B21,(*dw)->B21,len);
        cudaPutToGpu(_gpuData->B31,(*dw)->B31,len);
    }

    cudaCopyStrings(data,_cpuData->getSize().x);


    ////////////////////////////////////////////////////////////

    len=_cpuData->getSize().x;

    if(up)
    {
        cudaPutToGpu(_gpuData->Vs1,(*up)->Vs1,len);
        cudaPutToGpu(_gpuData->Sc1,(*up)->Sc1,len);
        cudaPutToGpu(_gpuData->Cs1,(*up)->Cs1,len);
        cudaPutToGpu(_gpuData->R11,(*up)->R11,len);
        cudaPutToGpu(_gpuData->R21,(*up)->R21,len);
        cudaPutToGpu(_gpuData->R31,(*up)->R31,len);
        cudaPutToGpu(_gpuData->G11,(*up)->G11,len);
        cudaPutToGpu(_gpuData->G21,(*up)->G21,len);
        cudaPutToGpu(_gpuData->G31,(*up)->G31,len);
        cudaPutToGpu(_gpuData->B11,(*up)->B11,len);
        cudaPutToGpu(_gpuData->B21,(*up)->B21,len);
        cudaPutToGpu(_gpuData->B31,(*up)->B31,len);
    }
    else
    {
        data.srcString=1;
        data.dstString=0;
        cudaCopyStrings(data,_cpuData->getSize().x);
    }
}

std::auto_ptr<GpuData> * Device::getUpMark()
{
    return &_upMark;
}

std::auto_ptr<GpuData> * Device::getDownMark()
{
    return &_dwMark;
}

int Device::getHeight()
{
    return _stringMarks.y-_stringMarks.x;
}

int Device::getGpuId()
{
    return _gpuId;
}
