#include "scenario.h"
#include "filemanager.h"
#include "device.h"
#include "log.h"

#include <map>

#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))

const char * Camera::CAMERA_POSITION_X="camera_position_x";
const char * Camera::CAMERA_POSITION_Y="camera_position_y";
const char * Camera::CAMERA_SIZE_X="camera_size_x";
const char * Camera::CAMERA_SIZE_Y="camera_size_y";
const char * Scenario::SCENARIO_FULL_SIZE_X="full_size_x";
const char * Scenario::SCENARIO_FULL_SIZE_Y="full_size_y";
const char * Scenario::SCENARIO_TOTAL_TIME="total_time";
const char * Scenario::SCENARIO_TOTAL_STEP="total_step";
const char * Scenario::SCENARIO_ENTERFRAME="enterframe";
const char * Scenario::SCENARIO_TIMEFRAME="timeframe";
const char * Scenario::SCENARIO_ENTERSAVE="entersave";
const char * Scenario::SCENARIO_OUTPUT_PATH="output_path";
const char * Scenario::SCENARIO_MAX_VISC="max_viscosity";
const char * Scenario::SCENARIO_SCHEME_CORREECTION="color_scheme_correction";

std::vector<std::vector<float> > Camera::getFrame(std::vector<std::vector<float> > & rgb, int xsize, int ysize)
{
    std::vector<std::vector<float> > out(3);

    out[0].resize(size.x*size.y);
    out[1].resize(size.x*size.y);
    out[2].resize(size.x*size.y);

    int length = xsize*ysize;
    int index = 0;

    for(int j=0;j<size.y;++j)
    {
        int y=j-size.y/2-position.y+ysize/2;
        int x=-size.x/2-position.x+xsize/2;
        float pos=(j-size.y/2-position.y+ysize/2)*xsize+
                (-size.x/2-position.x+xsize/2);
        for(int i=0;i<size.x;++i)
        {
            if(pos>=0&&pos<length&&x>=0&&y>=0&&x<xsize&&y<ysize)
            {
                out[0][index]=rgb[0][pos];
                out[1][index]=rgb[1][pos];
                out[2][index]=rgb[2][pos];
            }
            ++index;
            ++pos;
            ++x;
        }
    }

    return out;
}

std::map<std::string,UniPointer> Scenario::_getParams()
{
    std::map<std::string,UniPointer> out;
    out[GpuInfo::GPU_ID]=&(_gpuInfo[0].gpuId);
    out[Camera::CAMERA_SIZE_X]=&(_camera.size.x);
    out[Camera::CAMERA_SIZE_Y]=&(_camera.size.y);
    out[Camera::CAMERA_POSITION_X]=&(_camera.position.x);
    out[Camera::CAMERA_POSITION_Y]=&(_camera.position.y);
    out[Scenario::SCENARIO_ENTERFRAME]=&(_enterFrame);
    out[Scenario::SCENARIO_TIMEFRAME]=&(_timeFrame);
    out[Scenario::SCENARIO_ENTERSAVE]=&(_enterSave);
    out[Scenario::SCENARIO_FULL_SIZE_X]=&(_fullSize.x);
    out[Scenario::SCENARIO_FULL_SIZE_Y]=&(_fullSize.y);
    out[Scenario::SCENARIO_TOTAL_STEP]=&(_totalStep);
    out[Scenario::SCENARIO_TOTAL_TIME]=&(_totalTime);
    out[Scenario::SCENARIO_OUTPUT_PATH]=&(_outputPath);
    out[Scenario::SCENARIO_SCHEME_CORREECTION]=&(_schemeCorection);
    out[Scenario::SCENARIO_MAX_VISC]=&(_maxVisc);
    return out;
}

bool Scenario::_setParams(std::vector<std::string> input)
{
    std::map<std::string,UniPointer> setter=_getParams();

    for(std::vector<std::string>::iterator it=input.begin();it<input.end();++it)
    {
        std::vector<std::string> str = split_string(*it,'=');
        try
        {
            if(str.size()==2 && setter.count(str[0])>0)
            {
                if(str[1].size()<1)
                {
                    logger(std::string("no parameter in ")+str[0]);
                }
                if(str[0]==GpuInfo::GPU_ID)
                {
                    std::vector<std::string> fstr=split_string(str[1],',');
                    if(fstr.size()<1)
                    {
                        logger("no gpu ID in input");
                        return false;
                    }
                    _gpuInfo.resize(fstr.size());
                    for(int i=0;i<fstr.size();++i)
                    {
                        _gpuInfo[i].gpuId=stringToInt(fstr[i]);
                    }
                }
                else
                {
                    setter[str[0]]=str[1];
                    setter[str[0]]=stringToFloat(str[1]);
                }
            }
        }
        catch(...)
        {
            logger(std::string("exception in Scenario::_setParams"));
            return false;
        }
    }
    if(_fullSize.x<32)
    {
        logger(std::string("too small size x in Scenario::_setParams : " + intToString(_fullSize.x)));
        _fullSize.x=0;
        return false;
    }
    if(_fullSize.y<32)
    {
        logger(std::string("too small size y in Scenario::_setParams : " + intToString(_fullSize.y)));
        _fullSize.y=0;
        return false;
    }
    for(int i=0;i<_gpuInfo.size();++i)
        if(_gpuInfo[i].gpuId<0)
        {
            logger(std::string("wrong gpu ID in Scenario::_setParams : " + intToString(_gpuInfo[i].gpuId)));
            _gpuInfo[i].gpuId=0;
            return false;
        }
    if(_camera.size.x*_camera.size.y<=0)
    {
        logger(std::string("wrong camera size in Scenario::_setParams : " + intToString(_camera.size.x) + " x " + intToString(_camera.size.y)));
        _camera.size.x=_camera.size.y=1;
        return false;
    }
    if(_totalTime<=0)
    {
        logger(std::string("wrong total time in Scenario::_setParams : " + floatToString(_totalTime)));
        _totalTime=1;
        return false;
    }
    if(_totalStep<=0)
    {
        logger(std::string("wrong total step in Scenario::_setParams : " + floatToString(_totalStep)));
        _totalStep=1;
        return false;
    }
    if(_enterFrame<=0)
    {
        logger(std::string("wrong enterFrame distance in Scenario::_setParams : " + intToString(_enterFrame)));
        _enterFrame=1;
        return false;
    }
    if(_enterSave<=0)
    {
        logger(std::string("wrong enterFrame distance in Scenario::_setParams : " + intToString(_enterSave)));
        _enterSave=1;
        return false;
    }
    _fullSize.x=(_fullSize.x/32 + ((_fullSize.x%32)?1:0))*32;//
    _fullSize.y=(_fullSize.y/32 + ((_fullSize.y%32)?1:0))*32;//

    if(_outputPath.find_last_of('/')!=_outputPath.size()-1)
        _outputPath+="/";

    if(_outputPath.find("out/",0,4)!=_outputPath.size()-4)
        _outputPath+="out/";

    return true;
}

Scenario::Scenario(std::string pathToApplication)
{
    _totalTime=1e35f;
    _totalStep=2000000000;
    _enterFrame=50;
    _timeFrame=1.0;
    _enterSave=2000000000;
    _lastUpdateTime=-1;
   _fullSize.x=1920;
   _fullSize.y=1080;
   _schemeCorection=0;
   _sleepTime=0;
   _maxVisc=1e35;
   _outputPath=pathToApplication;
   _gpuInfo.push_back(GpuInfo());

    if(!_loadDefault(pathToApplication))
    {
        _saveDefault(pathToApplication);
    }
}

bool Scenario::_loadDefault(std::string pathToApplication)
{
    FileManager file;
    std::vector<std::string> input = file.read(pathToApplication+"default.txt");
    if(input.size()==0)
        return false;

    return _setParams(input);
}

bool Scenario::_saveDefault(std::string pathToApplication)
{
    FileManager file;
    std::map<std::string,UniPointer> getter=_getParams();
    std::vector<std::string> output;
    for(std::map<std::string,UniPointer>::iterator it = getter.begin();it != getter.end();++it)
    {
        if(it->first==GpuInfo::GPU_ID)
        {
            std::string str;
            for(int i=0;i<_gpuInfo.size();++i)
                str+=intToString(_gpuInfo[i].gpuId)+",";
            output.push_back(it->first + "=" + str);
            continue;
        }
        if(it->second.isInt())
        {
            output.push_back(it->first + "=" + intToString(it->second.getInt()));
        }
        else if(it->second.isFloat())
        {
            output.push_back(it->first + "=" + floatToString(it->second.getFloat()));
        }
        else if(it->second.isString())
        {
            output.push_back(it->first + "=" + it->second.getString());
        }
    }
    file.write(pathToApplication+"default.txt",output);
}

bool Scenario::init(std::string path)
{
    if(path.size()<1)
    {
        logger(std::string("wrong path in Scenario::init file : ")+path);
        return false;
    }
    _path=path;
    refresh(false);
    if(_commands.size()<1)
    {
        logger(std::string("no command in Scenario::init file : ")+path);
        return false;
    }
    if(!_setParams(_commands))
    {
        logger(std::string("can't set params in Scenario::init"));
        return false;
    }
    int totalMemory=0;
    for(int i=0;i<_gpuInfo.size();++i)
    {
        _gpuInfo[i]=Device::getGpuInfo(_gpuInfo[i].gpuId);
        totalMemory+=_gpuInfo[i].memory;
    }
    if(totalMemory<_fullSize.x*_fullSize.y)
    {
        logger(std::string("not enough gpu memory in Scenario::init avalible : "+
                        intToString(totalMemory) +
                        " ; trying to allocate : "+
                        intToString(_fullSize.x) + " x " + intToString(_fullSize.y) +
                        " = " + intToString(_fullSize.x*_fullSize.y)));
        return false;
    }
    return true;
}

bool Scenario::init(CpuData & cpuData)
{
    try
    {
        std::vector<std::vector<Gas> > gas;

        gas.resize(_fullSize.x);
        for(int i=0;i<_fullSize.x;++i)
            gas[i].resize(_fullSize.y);

        cpuData.init(gas);
        update(cpuData);
    }
    catch(...)
    {
        logger(std::string("exception in Scenario::init"));
        return false;
    }
    return true;
}

bool Scenario::update(CpuData & cpuData, std::vector<Device> * gpuData)
{
    if(cpuData.getTime()<_sleepTime)
        return true;

    _sleepTime=1e35f;

    bool isSynchronized=false;

    for(std::vector<std::string>::iterator it=_commands.begin();it<_commands.end();++it)
    {
        std::vector<std::string> str = split_string(*it,'=');
        try
        {
            if(str.size()==2)
            {
                std::vector<std::string> par=split_string(str[0],']');
                if(par.size()==2)
                {
                    par[0]=symbol_eraser(par[0],'[');
                    float time=stringToFloat(par[0]);
                    if(time<=_lastUpdateTime||cpuData.getTime()<time)
                    {
                        if(cpuData.getTime()<time)
                        {
                            _sleepTime=min(_sleepTime,time-1e-6f);
                        }
                        continue;
                    }
                    if(par[1].find("save",0,4)!=std::string::npos)
                    {
                        if(gpuData)
                        {
                            for(int i=0;i<gpuData->size();++i)
                            {
                                cpuData=&((*gpuData)[i]);
                            }
                        }
                        save(cpuData,str[1]);
                        if(par[1].find_first_of('q')!=std::string::npos)
                        {
                            _totalStep=-1;
                            _totalTime=-1;
                            return true;
                        }
                    }
                    else
                    {
                        std::vector<std::string> shiftstr=split_string(par[1],',');
                        if(shiftstr.size()!=2)
                        {
                            logger(std::string("wrong shift format in Scenario::update"));
                            return false;
                        }
                        shiftstr[0]=symbol_eraser(shiftstr[0].substr(4),'(');
                        shiftstr[1]=symbol_eraser(shiftstr[1],')');
                        Position shift(stringToFloat(shiftstr[0]),stringToFloat(shiftstr[1]));

                        if(str[1].find(".gas",0,4)==std::string::npos)
                        {
                            OutData outdata;
                            if(!outdata.open(str[1],shift))
                                continue;

                            if(gpuData&&!isSynchronized)
                                for(int d=0;d<gpuData->size();++d)
                                    cpuData=&(*gpuData)[d];

                            for(int i=cpuData.getZeroPosition().x;i<cpuData.getZeroPosition().x+cpuData.getSize().x;++i)
                            {
                                for(int j=cpuData.getZeroPosition().y;j<cpuData.getZeroPosition().y+cpuData.getSize().y;++j)
                                {
                                    cpuData.setGas(outdata.update(cpuData.getGas(Position(i,j)),Position(i,j)),Position(i,j));
                                }
                            }
                            std::vector<CpuGravityBody> body = outdata.getBodys(cpuData.getZeroPosition());
                            for(int i=0;i<body.size();++i)
                            {
                                cpuData.addBody(body[i]);
                            }
                        }
                        else
                        {
                            FileManager file;
                            std::vector<float> input = file.load(str[1]);
                            if(input.size()<1)
                                continue;
                            std::vector<std::vector<Gas> > save = deserilizeGas(input);
                            if(save.size()<1||save[0].size()<1)
                                continue;
                            if(gpuData&&!isSynchronized)
                                for(int d=0;d<gpuData->size();++d)
                                    cpuData=&(*gpuData)[d];

                            for(int i=cpuData.getZeroPosition().x;i<cpuData.getZeroPosition().x+cpuData.getSize().x;++i)
                            {
                                for(int j=cpuData.getZeroPosition().y;j<cpuData.getZeroPosition().y+cpuData.getSize().y;++j)
                                {
                                    Position pos(i-shift.x+save.size()/2,j-shift.y+save[0].size()/2);
                                    if(pos.x<0||pos.x>=save.size()||pos.y<0||pos.y>=save[0].size())
                                        continue;
                                    cpuData.setGas(save[pos.x][pos.y],Position(i,j));
                                }
                            }
                        }
                    }
                }
            }
        }
        catch(...)
        {
            logger(std::string("exception in Scenario::update"));
            return false;
        }
    }
    _lastUpdateTime=cpuData.getTime();

    if(gpuData&&isSynchronized)
        for(int d=0;d<gpuData->size();++d)
            if(!(*gpuData)[d].update())
                return false;

    return true;
}

void Scenario::refresh(bool getParam)
{
    FileManager file;
    std::vector<std::string> commands=file.read(_path);

    if(commands.size()==_commands.size())
    {
        int i=0;
        for(i=0;i<_commands.size();++i)
        {
            if(_commands[i]!=commands[i])
                break;
        }
        if(i==commands.size())
            return;
    }

    _sleepTime=0;
    _commands = commands;
    if(getParam)
        _setParams(_commands);
}

bool Scenario::inProgress(CpuData & cpuData)
{
    return (cpuData.getTime()<=_totalTime)&&(cpuData.getStep()<=_totalStep);
}

bool Scenario::makeVoid(std::string path, std::string firstElement)
{
    std::vector<std::string> out;

    std::map<std::string,UniPointer> params = _getParams();

    for(std::map<std::string,UniPointer>::iterator it = params.begin(); it != params.end(); ++it)
    {
        if(it->second.isInt())
        {
            out.push_back(it->first + "=" + intToString(it->second.getInt()));
        }
        else if(it->second.isFloat())
        {
            out.push_back(it->first + "=" + floatToString(it->second.getFloat()));
        }
        else if(it->second.isString())
        {
            out.push_back(it->first + "=" + it->second.getString());
        }
    }

    out.push_back(std::string(""));
    out.push_back(std::string("[0]load(0,0)=") + firstElement);
    out.push_back(std::string("//[10000.5]save(0,0)=autosave.gas"));

    FileManager file;
    return file.write(path,out);
}

float Scenario::getTotalTime()
{
    return _totalTime;
}

int Scenario::getTotalStep()
{
    return _totalStep;
}

int Scenario::getEnterFrame()
{
    return _enterFrame;
}

float Scenario::getTimeFrame()
{
    return _timeFrame;
}

int Scenario::getEnterSave()
{
    return _enterSave;
}

std::vector<GpuInfo> Scenario::getGpuInfo()
{
    return _gpuInfo;
}

Camera Scenario::getCamera()
{
    return _camera;
}

std::string Scenario::getOutputPath()
{
    return _outputPath;
}

std::string Scenario::getPath()
{
    return _path.substr(0,_path.find_last_of('/')+1);
}

float Scenario::getParamsSchemeCorrection()
{
    return _schemeCorection;
}

std::map<int,int> Scenario::buildGpuStrings(Position size)
{
    std::map<int,int> out;
    int stringBlocks=size.y/32;
    int totalMemory=0;
    for(int i=0; i<_gpuInfo.size();i++)
    {
        out[_gpuInfo[i].gpuId]=stringBlocks*_gpuInfo[i].memory;
        totalMemory+=_gpuInfo[i].memory;
    }
    for(int i=0; i<_gpuInfo.size();i++)
    {
        out[_gpuInfo[i].gpuId]/=totalMemory;
        stringBlocks-=out[_gpuInfo[i].gpuId];
    }
    while(stringBlocks>0)
    {
        int maxFreeMem=-1;
        int emptyDevice=-1000000;
        for(int i=0; i<_gpuInfo.size();i++)
        {
            if(_gpuInfo[i].memory/32-out[_gpuInfo[i].gpuId]>=maxFreeMem)
            {
                maxFreeMem=_gpuInfo[i].memory/32-out[_gpuInfo[i].gpuId];
                emptyDevice=_gpuInfo[i].gpuId;
            }
        }
        if(maxFreeMem>=0)
        {
            out[emptyDevice]++;
            stringBlocks--;
        }
        else
        {
            logger("can't find empty device in Scenario::buildGpuStrings");
        }
    }
    return out;
}

float Scenario::getMaxVisc()
{
    return _maxVisc;
}

bool Scenario::save(CpuData & cpuData, std::string path)
{
    FileManager file;
    file.save(path+".gas",serilizeGas(cpuData.getGas()));
    std::vector<std::string> commands;
    commands.push_back(std::string("[0]load(0,0)=")+path+".gas");
    for(std::vector<std::string>::iterator it=_commands.begin();it!=_commands.end();++it)
    {
        std::string command=*it;
        if(command.find_first_of('[')==std::string::npos)
        {
            commands.push_back(command);
            continue;
        }
        std::vector<std::string> fcom = split_string(command,'[');
        if(fcom.size()!=2)
            continue;
        fcom[0]=symbol_eraser(fcom[0],'[');
        fcom[0]=symbol_eraser(fcom[0],']');
        float time = stringToFloat(fcom[0]);
        if(time>cpuData.getTime())
        {
            commands.push_back(std::string("[")+floatToString(time-cpuData.getTime())+"]="+fcom[1]);
        }
    }
    file.write(path+".txt",commands);
}
